package jp2codec.imageio;

import javax.imageio.ImageReader;
import javax.imageio.ImageReadParam;
import javax.imageio.IIOException;
import javax.imageio.ImageTypeSpecifier;
import javax.imageio.stream.ImageInputStream;
import javax.imageio.spi.ImageReaderSpi;
import javax.imageio.metadata.IIOMetadata;
import java.io.IOException;
import java.io.EOFException;
import java.util.Iterator;
import java.util.ArrayList;
import java.awt.image.*;
import java.awt.color.ColorSpace;
import java.awt.*;
import java.nio.ByteOrder;


import jp2codec.quantization.dequantizer.*;
import jp2codec.image.invcomptransf.*;
import jp2codec.fileformat.reader.*;
import jp2codec.codestream.reader.*;
import jp2codec.wavelet.synthesis.*;
import jp2codec.entropy.decoder.*;
import jp2codec.image.*;
import jp2codec.roi.*;
import jp2codec.decoder.DecoderSpecs;

/**
 * This class is responsible for responding to queries about the images actually
 * stored in an input file or stream, as well as the actual reading of images,
 * thumbnails, and metadata. For simplicity, we will ignore thumbnail images in
 * this class.
 */
public class JP2ImageReader extends ImageReader {

    ImageInputStream in = null;
    int width, height;
    int colorType;
    WritableRaster imRas;
    BitstreamReaderAgent breader;

    // Constants enumerating the values of colorType
    static public final int COLOR_TYPE_GRAY = 0;
    static public final int COLOR_TYPE_RGB = 1;
    static public final int COLOR_TYPE_RGBA = 2;

    private HeaderDecoder hd;

    /**
     * Indicates whether file header was readed or not
     */
    boolean gotHeader = false;

    /**
     * The constructor of JP2ImageReader
     */
    public JP2ImageReader(ImageReaderSpi originatingProvider) {
        super(originatingProvider);
    }

    /**
     * Returns 1 since multi-image files are not supported
     *
     * @param allowSearch
     * @return 1
     * @throws IOException
     */
    public int getNumImages(boolean allowSearch) throws IOException {
        return 1;
    }

    /**
     * Checks if image index is equal 0 since multi-image files are not supported
     *
     * @param imageIndex The image index
     */
    private void checkIndex(int imageIndex) {
        if (imageIndex != 0) {
            throw new IndexOutOfBoundsException("bad index");
        }
    }

        /**
     * Returns width of the image
     *
     * @param imageIndex The image index
     * @return Width of the image
     * @throws IOException
     */
    public int getWidth(int imageIndex)
            throws IOException {
        checkIndex(imageIndex); // must throw an exception if != 0
        readHeader();
        return hd.getImgWidth();
    }

    /**
     * Returns height of the image
     *
     * @param imageIndex The image index
     * @return Width of the image
     * @throws IOException
     */
    public int getHeight(int imageIndex)
            throws IOException {
        checkIndex(imageIndex);
        readHeader();
        return hd.getImgHeight();
    }

    public IIOMetadata getStreamMetadata() throws IOException {
        return null;
    }

    public String getFormatName() throws IOException {
        return "JPEG2000";
    }

    public IIOMetadata getImageMetadata(int imageIndex) throws IOException {
        return null;
    }

    public Iterator<ImageTypeSpecifier> getImageTypes(int imageIndex)
            throws IOException {
        checkIndex(imageIndex);
        readHeader();

        ImageTypeSpecifier imageType = null;
        int datatype = DataBuffer.TYPE_BYTE;
        java.util.List<ImageTypeSpecifier> l = new ArrayList<ImageTypeSpecifier>();
        switch (colorType) {
            case COLOR_TYPE_GRAY:
                imageType = ImageTypeSpecifier.createGrayscale(8,
                        datatype,
                        false);
                break;

            case COLOR_TYPE_RGB:
                ColorSpace rgb =
                        ColorSpace.getInstance(ColorSpace.CS_sRGB);
                int[] bandOffsets = new int[3];
                bandOffsets[0] = 0;
                bandOffsets[1] = 1;
                bandOffsets[2] = 2;
                imageType =
                        ImageTypeSpecifier.createInterleaved(rgb,
                                bandOffsets,
                                datatype,
                                false,
                                false);
                break;
        }
        l.add(imageType);
        return l.iterator();
    }

    /**
     * This method  is responsible for reading enough of the input stream
     * to determine the width, height, and structure of the image. readHeader
     * is defined so it is safe to be called multiple times.
     *
     * @throws IOException
     */
    private void readHeader() throws IOException {
        if (gotHeader) {
            return;
        }

        synchronized ((Object) gotHeader) {

            if (input == null) {
                throw new IllegalStateException("No input stream");
            }

            in = (ImageInputStream) input;
            in.setByteOrder(ByteOrder.BIG_ENDIAN);

            // **** File Format ****
            // If the codestream is wrapped in the jp2 fileformat, Read the
            // file format wrapper

            FileFormatReader ff = new FileFormatReader(in);
            ff.readFileFormat();
            if (ff.JP2FFUsed) {
                in.seek(ff.getFirstCodeStreamPos());
            }

            try {
                hd = new HeaderDecoder(in);
            } catch (EOFException e) {
                throw new IIOException("Codestream too short or bad header.");
            }
            gotHeader = true;
        }
    }

    public void setROI(int x, int y, int w, int h) {
        if (breader != null) breader.setROI(x, y, w, h);
    }

    public BufferedImage read(int imageIndex, ImageReadParam param) throws IOException {

        clearAbortRequest();
        // **** Header decoder ****
        // Instantiate header decoder and read main header
        readHeader();
        DecoderSpecs decSpec = hd.getDecoderSpecs();
        if (abortRequested()) return null;

        // **** Instantiate decoding chain ****
        int nComp; // number of components in the image
        int depth[]; // Get demixed bitdepths
        nComp = hd.getNumComps();
        depth = new int[nComp];
        for (int i = 0; i < nComp; i++) {
            depth[i] = hd.getOriginalBitDepth(i);
        }

        // **** Bitstream reader ****
        try {
            breader = new BitstreamReaderAgent(hd, in, decSpec, -1, -1, false, -1);
        } catch (IOException e) {
            throw new IIOException("Error while reading bit stream header or parsing " +
                    "packets" + ((e.getMessage() != null) ?
                    (":\n" + e.getMessage()) : ""));
        } catch (IllegalArgumentException e) {
            throw new IIOException("Could not instantiate bit stream reader" +
                    ((e.getMessage() != null) ?
                    (":\n" + e.getMessage()) : ""));
        }

        // **** Entropy decoder ****
        EntropyDecoder entdec;
        try {
            entdec = new StdEntropyDecoder(breader, decSpec, true, true); // true = Cer + Cverber
        } catch (IllegalArgumentException e) {
            throw new IIOException("Could not instantiate entropy decoder" +
                    ((e.getMessage() != null) ?
                    (":\n" + e.getMessage()) : ""));
        }

        // **** ROI de-scaler ****
        ROIDeScaler roids;
        try {
            roids = new ROIDeScaler(entdec, decSpec.rois);
        } catch (IllegalArgumentException e) {
            throw new IIOException("Could not instantiate roi de-scaler." +
                    ((e.getMessage() != null) ?
                    (":\n" + e.getMessage()) : ""));
        }

        // **** Dequantizer ****
        Dequantizer deq;
        try {
            deq = new StdDequantizer(roids, depth, decSpec);
        } catch (IllegalArgumentException e) {
            throw new IIOException("Could not instantiate dequantizer" +
                    ((e.getMessage() != null) ?
                    (":\n" + e.getMessage()) : ""));
        }

        // **** Inverse wavelet transform ***
        InverseWT invWT;
        try {
            // full page inverse wavelet transform
            invWT = new InverseWT(deq, decSpec);
        } catch (IllegalArgumentException e) {
            throw new IIOException("Could not instantiate inverse wavelet transform" +
                    ((e.getMessage() != null) ?
                    (":\n" + e.getMessage()) : ""));
        }

        int res; // resolution level to reconstruct
        res = breader.getImgRes();
        invWT.setImgResLevel(res);

        // **** Data converter **** (after inverse transform module)
        ImgDataConverter converter;
        converter = new ImgDataConverter(invWT, 0);

        // **** Inverse component transformation ****
        InvCompTransf ictransf;
        InvCompTransf src = ictransf = new InvCompTransf(converter, decSpec, depth);

        // -------------------------------------------------------------------------------

        width = ictransf.getImgWidth();
        height = ictransf.getImgHeight();

        // Check for image type
        switch (src.getNumComps()) {
            case 1:
                colorType = COLOR_TYPE_GRAY;
                break;
            case 3:
                colorType = COLOR_TYPE_RGB;
                break;
            case 4:
                colorType = COLOR_TYPE_RGBA;
                break;
            default:
                throw new IllegalArgumentException("Only 1, 3, and 4 components " +
                        "supported");
        }

        // Check component sizes and bitdepths
        for (int i = src.getNumComps() - 1; i >= 0; i--) {
            if (src.getCompImgHeight(i) != src.getImgHeight() ||
                    src.getCompImgWidth(i) != src.getImgWidth()) {
                throw new IllegalArgumentException("All components must have " +
                        "the same dimensions and no" +
                        "subsampling");
            }
            if (src.getNomRangeBits(i) > 8) {
                throw new IllegalArgumentException("Depths greater than 8 " +
                        "bits per component is not " +
                        "supported");
            }
        }
        //-------------------------------------------------------------------------

        // Compute initial source region, clip against destination later
        Rectangle sourceRegion = getSourceRegion(param, width, height);

        int inputBands;
        switch (colorType) {
            case COLOR_TYPE_RGBA:
                inputBands = 4;
                break;
            case COLOR_TYPE_RGB:
                inputBands = 3;
                break;
            default:
                inputBands = 1;
        }

        // Set everything to default values
        int sourceXSubsampling = 1;
        int sourceYSubsampling = 1;
        int[] sourceBands = {0, 1, 2};
        int[] destinationBands = new int[inputBands];
        for (int i = 0; i < inputBands; i++) destinationBands[i] = i;
        Point destinationOffset = new Point(0, 0);

        // Get values from the ImageReadParam, if any
        if (param != null) {
            sourceXSubsampling = param.getSourceXSubsampling();
            sourceYSubsampling = param.getSourceYSubsampling();
            sourceBands = param.getSourceBands();
            destinationBands = param.getDestinationBands();
            destinationOffset = param.getDestinationOffset();
        }

        // Get the specified detination image or create a new one
        BufferedImage dst = getDestination(param,
                getImageTypes(0),
                width, height);

        // Enure band settings from param are compatible with images
        checkReadParamBandSettings(param, inputBands,
                dst.getSampleModel().getNumBands());

        int[] bandOffsets = new int[inputBands];
        for (int i = 0; i < inputBands; i++) {
            bandOffsets[i] = i;
        }

        int bytesPerRow = width * inputBands;
        DataBufferByte rowDB = new DataBufferByte(bytesPerRow);
        WritableRaster rowRas =
                Raster.createInterleavedRaster(rowDB,
                        width, 1, bytesPerRow,
                        inputBands, bandOffsets,
                        new Point(0, 0));
        byte[] rowBuf = rowDB.getData();

        // Create an int[] that can a single pixel (0,0)
        int[] pixel = rowRas.getPixel(0, 0, (int[]) null);

        imRas = dst.getWritableTile(0, 0);
        int dstMinX = imRas.getMinX();
        int dstMaxX = dstMinX + imRas.getWidth() - 1;
        int dstMinY = imRas.getMinY();
        int dstMaxY = dstMinY + imRas.getHeight() - 1;

        Coord nT = src.getNumTiles(null);

        // Start the data delivery to the cached consumers tile by tile
        for (int y = 0; y < nT.y; y++) {
            for (int x = 0; x < nT.x; x++) {
                src.setTile(x, y);

                // Initialize tile
                int height = src.getHeight();   // tile height
                int width = src.getWidth();     // tile width

                // The offset of the active tiles is the same for all
                // components,
                // since we don't support different component dimensions.
                int tOffx = src.getULX(0) -
                        (src.getImgULX() + src.getCompSubsX(0) - 1) / src.getCompSubsX(0);
                int tOffy = src.getULY(0) -
                        (src.getImgULY() + src.getCompSubsY(0) - 1) / src.getCompSubsY(0);

                for (int j : destinationBands) {
                    for (int yy = 0; yy < height; yy++)
                        for (int xx = 0; xx < width; xx++) {
                            imRas.setSample(xx, yy, j, 128);
                        }
                }

                while (breader.anyCodeBlockRemains()) {
                    if (abortRequested()) return dst;
                    src.invCT(dst);
                    processImageUpdate(dst, 0, 0, dst.getWidth(), dst.getHeight(), 1, 1, null);
                    processImageProgress((float) breader.cbUsed / breader.pktDec.cbIdx * 100);
                }
            } // End loop on horizontal tiles
        } // End loop on vertical tiles
        processImageComplete();

        return dst;
    }

}

